
;=== Declaration des variables globales

;= la liste des regles
(defvar *regles*)
;= la liste des faits
(defvar *faits*)
;= variable globale permettant de savoir combien de regles ont ete appliquees durant un tour d'inferences
(defvar *nb_regles_appl*)
;= variable permettant de detecter des defauts de bouclage (fait modifie plusieurs fois)
(defvar *risque_boucle*)
;= compteur des tours d'inferences
(defvar *nb_tours_inf*)


;=== Implementation des regles

(defun regles_init ( )
  (progn
    (setq *regles* '(
       ( R1  ( (1 poils)                                                             ) ( (1 mammifere) ) )
       ( R2  ( (1 donne-lait)                                                        ) ( (1 mammifere) ) )
       ( R3  ( (1 plumes)                                                            ) ( (1 oiseau)    ) )
       ( R4  ( (1 vole) (1 ponds-oeufs)                                              ) ( (1 oiseau)    ) )
       ( R5  ( (1 mange-viande)                                                      ) ( (1 carnivore) ) )
       ( R6  ( (1 dents-pointues) (1 griffes) (1 yeux-vers-avant)                    ) ( (1 carnivore) ) )
       ( R7  ( (1 mammifere) (1 sabots)                                              ) ( (1 ongule)    ) )
       ( R8  ( (1 mammifere) (1 rumine)                                              ) ( (1 ongule)    ) )
       ( R9  ( (1 mammifere) (1 carnivore) (1 brun) (1 taches)                       ) ( (1 guepard)   ) )
       ( R10 ( (1 mammifere) (1 carnivore) (1 brun) (1 raies)                        ) ( (1 tigre)     ) )
       ( R11 ( (1 ongule) (1 long-cou) (1 longues-pattes) (1 taches)                 ) ( (1 girafe)    ) )
       ( R12 ( (1 ongule) (1 raies)                                                  ) ( (1 zebre)     ) )
       ( R13 ( (1 oiseau) (0 vole) (1 long-cou) (1 longues-pattes) (1 noir-et-blanc) ) ( (1 autruche)  ) )
       ( R14 ( (1 oiseau) (0 vole) (1 noir-et-blanc)                                 ) ( (1 pingouin)  ) )
       ( R15 ( (1 oiseau) (1 vole)                                                   ) ( (1 albatros)  ) )
                    )
    )
    (regles_init_plist 0)
	 (setq *nb_regles_appl* 0)
	 (setq *risque_boucle* 0)
    (setq *nb_tours_inf* 1)
  )
)

(defun regles_init_plist (index)
  (if (nth index *regles*)
    (progn
      (putprop (car (nth index *regles*)) 0 'marquee)
      (regles_init_plist (+ index 1))
    )
    T
  )
)


;=== Implementation des faits initiaux : creation puis initialisation de leur valeur a 1 (true)

(defun faits_init ( )
  (progn
    ;== Pour le zebre
    (setq *faits* '( poils sabots raies ) )
    (faits_init_plist 0)
  )
)

(defun faits_init_plist (index)
  (if (nth index *faits*)
    (progn
      (putprop (nth index *faits*) 1 'valeur)
      (putprop (nth index *faits*) 0 'modif)
      (faits_init_plist (+ index 1))
    )
    T
  )
)


;=== fonctions diverses

;= permet de recuperer le numero-ieme element d'une liste (index de depart : 0)
(defun nth (numero liste)
  (if liste
    (if (<= numero 0)
      (car liste)
      (nth (- numero 1) (cdr liste))
    )
    nil
  )
)

(defun member (element liste)
  (if (position element liste)
    T
    nil
  )
)

;= index d'un element dans une liste : retour = [0..n] ou nil si absent
(defun position (element liste)
  (position_subr element liste 0)
)

(defun position_subr (element liste index)
  (if liste
    (if (equal element (car liste))
      index
      (position_subr element (cdr liste) (+ index 1))
    )
    nil
  )
)

; ******************************************************************************
; * MOTEUR D'INFERENCES :                                                      *
; *  - boucle_regles : Cette fonction commande un tour d'inference sur toutes  *
; * les regles. Elle se charge de gerer le 'nettoyage' des p-list des faits et *
; * des regles (faits modifies et regles appliquees au cours du balayage)      *
; * avant de relancer un autre tour d'inferences.                              *
; *  - boucle_regles_subr : Cette fonction passe la liste des regles en revue  *
; * et tente de les appliquer les unes apres les autres. Une fois la liste     *
; * entierement parcourue, elle repasse la main a boucle_regles.               *
; *  - boucle_premisses : pour chaque regle, cette fonction est appelee. Elle  *
; * passe les premisses en revue et renvoie T si la regle est applicable.      *
; *  - renvoi_valeur_fait : teste si le fait passe en parametre est present    *
; * la liste de faits et renvoie sa valeur (issue de sa p-list) ou nil.        *
; ******************************************************************************


(defun boucle_regles ()
  (progn
    (princ"\nParcours des regles : Tour n")
    (princ *nb_tours_inf*)
    (princ".\n")

    (nettoyage_faits 0)
;    (nettoyage_regles 0)
    (setq *nb_regles_appl* 0)
    (boucle_regles_subr *regles*)
    (setq *nb_tours_inf* (+ 1 *nb_tours_inf*))

    (princ"\nLes regles ont toutes ete passees une fois : ")
    (princ *nb_regles_appl*)
    (princ" ont ete appliquees.\n")
    (princ"Faits : ")
    (affichage_faits 0)
    (princ"\n")

    (if (equal *nb_regles_appl* 0)
      (progn
        (princ"\nFin : plus aucune regle n'est applicable.")
        T
      )
      (progn
        (if (equal *risque_boucle* 1)
          (princ"\nATTENTION : Au moins un fait a ete modifie plusieurs fois, risques de bouclage.")
          T
        )
        (princ"\n  Voulez-vous continuer (0=oui, 1=non) ? ")
        (setq choixutil (read))
        (if (equal choixutil 0)
          (boucle_regles)
          T 
        )
      )
    )
  )
)


(defun boucle_regles_subr (listeregles)
  ;== s'il reste des regles a tester
  (if listeregles
    (progn
      ;== Si la regle n'est pas marquee
      (if (equal (get (car (car listeregles)) 'marquee) 0)
        ;== Si ses premisses sont valides
        (if (boucle_premisses (car (cdr (car listeregles))))
          (progn
            (setq *nb_regles_appl* (+ *nb_regles_appl* 1))
            (princ"   Regle applicable : ")
            (princ(car (cdr (car listeregles))))
            (princ" --> ")
            ;== Mise a 1 du flag de la regle appliquee
            (putprop (car (car listeregles)) 1 'marquee)
            ;== Lancement de AppliqueRegle (modification des faits et tests des faits deja modifies)
				(applique_regle (car (cdr (cdr (car listeregles)))))
            (princ"\n")
          )
          nil
        )
        nil
      )
      (boucle_regles_subr (cdr listeregles))
    )
    ;== S'il ne reste plus de regles, on retourne T
    T
  )
)

(defun applique_regle (partiegauche)
  ;== S'il reste des faits a appliquer
  (if partiegauche
    (progn
      (let* ( (faittraite (car (cdr (car partiegauche)))) (valeurfait (car (car partiegauche))) )
        (if (renvoi_valeur_fait faittraite)
          ;== Si le fait existe dans la liste de faits
          (if (equal (get faittraite 'modif) 1)
            ;== Si le fait a deja ete modifie
            (progn
              (setq *risque_boucle* 1)
              (putprop faittraite valeurfait 'valeur)
              (putprop faittraite 1 'modif)
              (princ faittraite)
              (princ"=")
              (princ valeurfait)
              (princ" [RE-MODIF], ")
            )
            ;== Si le fait n'a pas encore ete modifie
            (progn
              (putprop faittraite valeurfait 'valeur)
              (putprop faittraite 1 'modif)
              (princ faittraite)
              (princ"=")
              (princ valeurfait)
              (princ" [MODIF], ")
            )
          )
          ;== Si le fait doit etre ajoute a la liste de faits
          (progn
            (setq *faits* (cons faittraite *faits*))
            (putprop faittraite valeurfait 'valeur)
            (putprop faittraite 0 'modif)
            (princ faittraite)
            (princ"=")
            (princ valeurfait)
            (princ" [CREATION], ")
          )
        )
      )
      (applique_regle (cdr partiegauche))
    )
    ;== S'il ne reste plus de faits a appliquer
    T
  )
)


(defun boucle_premisses (premisses)
  (if premisses
    (if (equal (car (car premisses)) (renvoi_valeur_fait (car (cdr (car premisses)))))
      (boucle_premisses (cdr premisses))
      nil
    )
    T
  )
)


; ***** Petites fonctionnalites du moteur d'inferences *****


(defun renvoi_valeur_fait (fait)
  (if (member fait *faits*)
    (get fait 'valeur)
    nil
  )
)

(defun nettoyage_regles (index)
  (if (nth index *regles*)
    (progn
      (putprop (car (nth index *regles*)) 0 'marquee)
      (nettoyage_regles (+ index 1))
    )
    T
  )
)

(defun nettoyage_faits (index)
  (if (nth index *faits*)
    (progn
      (putprop (nth index *faits*) 0 'modif)
      (nettoyage_faits (+ index 1))
    )
    T
  )
)

(defun affichage_faits (index)
  (if (nth index *faits*)
    (progn
      (princ (nth index *faits*))
      (princ"=")
      (princ (get (nth index *faits*) 'valeur))
      (princ", ")
      (affichage_faits (+ index 1))
    )
    T
  )
)


(regles_init)
(faits_init)
(boucle_regles)
