/*************************************************/
/* TP PROLOG : SATISFACTION DE CONTRAINTES       */
/*                                               */ 
/* Binome : Aurelien COGOLUEGNES / Thomas MANGIN */
/*                                               */
/* Date : 19 mars 1999                           */
/*************************************************/


/* Commande Permettant de lancer les tests */
/* */
/* inidom(11, Ldom) , statistics(cputime, T1) , statistics(inferences, I1) , backtrack2(Ldom, Laff) , statistics(cputime, T2) , statistics(inferences, I2) , TPS is (T2-T1) , INF is (I2-I1) , write(TPS) , nl , write(INF) , nl , ! , fail. */
/* */



/* Predicats du cours et necessaires pour ce TP */

aplist([], L, L)            .
aplist([X|L1], L2, [X|L3])  :- aplist(L1, L2, L3).

renverse([], [])      .
renverse([X|L1], L2)  :- renverse(L1, L3), aplist(L3, [X], L2).

appart(Nom, [Nom|_])   .
appart(Nom, [_|Reste]) :- appart(Nom, Reste).



/* Exercice 1.1 : listent(+Nombre, -Liste) */
/* Constitution d'une liste de N elements : [1, 2, ..., N] */

listentsubr(1, B, [B])   .  /* Condition d'arret : on a decrementer A de N jusqu'a 1 */
listentsubr(A, B, [B|L]) :- 
 C is (A-1)              ,  /* On decremente A : Descende avec A de N a 1 */
 D is (B+1)              ,  /* B est incremente : Remontee avec B de 1 a N */
 listentsubr(C, D, L)    .  /* On descend au niveau inferieur (suite de la liste) */

listent(0, [])           .  /* Si on demande une liste nulle */
listent(N, L)            :- 
 listentsubr(N, 1, L)    .  /* Lancement de la routine avec les deux compteurs inverses */



/* Exercice 1.2 : inidom(+Nombre, -ListeDom) */
/* Constitution d'une liste de N domaines : [domaine(X1, D1), ... domaine(XN, DN)] */

inidomsubr(N1, 1, [domaine(1, DI)])        :- /* Condition d'arret : on a atteint 1 */
 listent(N1, DI)                           .  /* On cree un premier domaine */
inidomsubr(N1, N2, [domaine(N2, DI)|Ldom]) :- /* On ajoute un domaine avec DI implemente a la remontee */
 M is (N2-1)                               ,  /* M sert de compteur de profondeur */
 inidomsubr(N1, M, Ldom)                   ,  /* On descend au niveau inferieur (suite de la liste) */
 listent(N1, DI)                           .  /* On implemente DI avant la remontee en utilisant N1 */

inidom(N, Ldom)                            :- /* On doit garder N pour l'implementation des domaines, */
 inidomsubr(N, N, Ldom2)                   ,  /* on lance donc la sous-routine avec un 2eme compteur */
 renverse(Ldom2, Ldom)                     .



/* Exercice 1.3 : valide(affect(Xi, Vi), affect(Xj, Vj)) */
/* Test de la validite de l'affectation de deux valeurs */

valide(affect(XI, VI), affect(XJ, VJ)) :-
 XI =\= XJ                             ,  /* Pas sur la meme ligne */
 VI =\= VJ                             ,  /* Pas sur la meme colonne */
 abs(XI-XJ) =\= abs(VI-VJ)             .  /* Pas sur la meme diagonale */



/* Exercice 2.1 : consistant(affect(Xi, Vi), Laff) */
/* Test de la validite d'une affectation par rapport a une liste d'affectations deja faites */

consistant(_, [])                                 .   /* Condition d'arret */
consistant(affect(XI, VI), [affect(XJ, VJ)|Laff]) :-  /* On teste le Car de la liste */ 
 valide(affect(XI, VI), affect(XJ, VJ))           ,   /* S'il est valide */
 consistant(affect(XI, VI), Laff)                 .   /* On continue sur le suivant */



/* Exercice 2.2 : backtrack1(+Ldom, -Laff) ou backtrack2(+Ldom, -Laff) */
/* Backtrack basique */

btsubr([], L, L)                            .   /* Condition d'arret : unification de la liste de retour */
btsubr([domaine(XI, DI)|Ldom], Laff, Laffr) :-  /* On recupere la variable suivante */
 appart(VI, DI)                             ,   /* On prend une valeur dans le domaine */
 once(consistant(affect(XI, VI), Laff))     ,   /* On teste si elle est consistante (mais pas a la remontee) */
 btsubr(Ldom, [affect(XI, VI)|Laff], Laffr) .   /* On boucle sur la variable suivante */
 
backtrack1(Ldom, Laff)     :-   /* On passe un Laff vide, on obtient 1 SOLUTION */
 btsubr(Ldom, [], Laff2)   ,    /* Appel de la subroutine : on part d'une Laff vide pour en avoir une pleine */
 reverse(Laff2, Laff)      .    /* Une fois la liste des affectations reue, on la renverse */

backtrack2(Ldom, SOLUCES)                       :-  /* On obtient TOUTES les solutions */
 findall(Laff, backtrack1(Ldom, Laff), SOLUCES) .



/* Exercice 3.1 : btfiltre1(+Ldom, -Laff) ou btfiltre2(+Ldom, -Laff) */
/* Backtrack avec forward-checking (filtrage des domaines rendus inconsistants) */

btfsubr([], L, L)                             .  /* Condition d'arret similaire */
btfsubr([domaine(XI, DI)|Ldom], Laff, Laffr)  :- /* On recupere le domaine(X, V) suivant */
 appart(VI, DI)                               ,  /* On prend une valeur : ELLE EST FORCEMENT CONSISTANTE */
 once(filtrage(XI, VI, Ldom, LdomR))          ,  /* On appelle le forward-checking (pas a la remontee) */
 btfsubr(LdomR, [affect(XI, VI)|Laff], Laffr) .  /* On boucle sur la variable suivante */

btfiltre1(Ldom, Laff)     :-  /* On passe un Laff vide, on obtient 1 SOLUTION */
 btfsubr(Ldom, [], Laff2) ,   /* Appel de la sub-routine */
 reverse(Laff2, Laff)     .   /* Inversion de la reponse (lisibilite facilite) */

btfiltre2(Ldom, SOLUCES)                       :-  /* Commande pour obtenir TOUTES les solutions */
 findall(Laff, btfiltre1(Ldom, Laff), SOLUCES) .

/* Ces routines de filtrage sont utilisees par les deux derniers backtrack */

filtragesubr(X, V, XI, DI, DISOL)     :-  /* Sous-routine de filtrage */
 appart(VI, DI)                       ,   /* Elle renvoie une solution seulement */
 valide(affect(X, V), affect(XI, VI)) ,   /* Il faut donc l'appeler avec un findall */
 DISOL is VI                          .   /* Renvoie de la solution */

filtrage(_, _, [], [])                                              .  /* Arret du filtrage : unification des listes */
filtrage(_, _, [domaine(0, 0)], [])                                 .  /* Arret du filtrage : backtrack mini (le suivant) */
filtrage(X, V, [domaine(XI, DI)|Ldom], [domaine(XI, DISOL2)|LdomR]) :- /* On recupere les domaines */
 findall(DISOL, filtragesubr(X, V, XI, DI, DISOL), DISOL2)          ,  /* On cherche toutes les solutions */
 filtrage(X, V, Ldom, LdomR)                                        .  /* On boucle sur le domaine suivant */



/* Exercice 4 : btfmini1(+Ldom, -Laff) ou btfmini2(+Ldom, -Laff) */
/* Backtrack avec filtrage et selection du domaine minimum */

btfmsubr([], L, L)                                    .   /* Condition d'arret */
btfmsubr([domaine(XI, DI)|Ldom], Laff, Laffr)         :-  /* On recupere le domaine suivant */
 once(select_dom(Ldom, domaine(X, D), Ldom2))         ,   /* On envoie la recherche du plus petit domaine */
 appart(VI, DI)                                       ,   /* Tests d'affectations */
 once(filtrage(XI, VI, [domaine(X, D)|Ldom2], LdomR)) ,   /* On lance le forward-checking */
 btfmsubr(LdomR, [affect(XI, VI)|Laff], Laffr)        .   /* On boucle sur le domaine suivant */

btfmini1(Ldom, Laff)       :-  /* Toujours pareil : lancement pour obtenir 1 SOLUTION */
 btfmsubr(Ldom, [], Laff2) ,
 reverse(Laff2, Laff)      .

btfmini2(Ldom, SOLUCES)                       :-  /* Lancement pour obtenir TOUTES les solutions */
 findall(Laff, btfmini1(Ldom, Laff), SOLUCES) .


/* Routines pour la selection du plus petit domaine */

/* Pour obtenir une liste contenant les longueurs de chaque domaine */
comptelong([], [])                           .   /* Condition d'arret */
comptelong([domaine(_, DI)|Ldom], [LG|LONG]) :-  /* On prend le domaine suivant */
 length(DI, LG)                              ,   /* On prend sa taille */
 not (LG =:= 0)                              ,   /* Si elle est differente de 0 */
 comptelong(Ldom, LONG)                      .   /* On unifie dans la liste de retour et on boucle */

/* Pour trouver l'index de la plus petite valeur (on doit utiliser un appart) */
chercheval(LONGUEURS, VAL, INDEX) :-  /* Val est la valeur a rechercher : on commence a 0, puis 1 ... */
 appart(D, LONGUEURS)             ,   /* On parcours toutes les valeurs dans le tableau */
 D =:= VAL                        ,   /* S'il y en a une qui correspond */
 nth1(INDEX, LONGUEURS, D)        .   /* On renvoie son index */

/* Routine principale pour la selection du plus petit domaine */
select_dom([], domaine(0, 0), [])                .  /* Condition d'arret : le domaine trouve est (O,0). */
select_dom(Ldom, domaine(X, D), L)               :- /* On recoie la liste des domaine et les var de retour */
 comptelong(Ldom, TABL)                          ,  /* On cree une liste contenant les longeurs des domaines */
 appart(VAL, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]) ,  /* Valeurs possibles du minimum */
 chercheval(TABL, VAL, INDEX)                    ,  /* On recherche le minimum */
 nth1(INDEX, Ldom, domaine(X, D))                ,  /* Son index dans la liste des longueurs est celui du domaine dans Ldom */
 select(Ldom, domaine(X, D), L)                  .  /* On enleve le domaine de Ldom et on renvoie les deux separemment */


